"""Работа с компонентами Windows"""
import os
import subprocess
import win32com.client
from .core import ms
from .path import PATH_TYPES, path2str
from typing import *
shell: dict[str, win32com.client.CDispatch] = {}
shell["WScript"] = win32com.client.Dispatch("WScript.Shell")
_names = {}
_names["lnk"] = {"args": "Arguments", "cwd": "WorkingDirectory", "desc": "Description", "hotkey": "Hotkey", "icon": "IconLocation", "lnk": "FullName", "target": "TargetPath"}


def read_lnk(path: PATH_TYPES) -> dict:
  """Прочитать ярлык .lnk"""
  r = {}
  lnk = shell["WScript"].CreateShortcut(path2str(path, True).replace("/", "\\"))
  r["src"] = lnk
  for k, v in _names["lnk"].items():
    r[k] = getattr(lnk, v, None)
  return r


def write_lnk(path: PATH_TYPES, target: str, args: str = None, cwd: str = None, desc: str = None, hotkey: str = None, icon: str = None):
  """Создать ярлык .lnk"""
  lnk = shell["WScript"].CreateShortCut(path2str(path, True).replace("/", "\\"))
  lnk.TargetPath = target.replace("/", "\\")
  if args != None:
    lnk.Arguments = args
  if cwd != None:
    lnk.WorkingDirectory = cwd.replace("/", "\\")
  if desc != None:
    lnk.Description = desc
  if hotkey != None:
    lnk.Hotkey = hotkey
  if icon != None:
    lnk.IconLocation = icon.replace("/", "\\")
  lnk.Save()
  return lnk


def hide_file(path: PATH_TYPES, *, recursive: bool = False, unhide: bool = False):
  """Скрыть файл используя системную команду `attrib`"""
  args = ["attrib"]
  args.append("-H" if unhide else "+H")
  args.append(path2str(path, True).replace("/", "\\"))
  if recursive:
    args += ["/S", "/D"]
  subprocess.run(args, check=True)


def unhide_file(path: PATH_TYPES, *, recursive: bool = False, hide: bool = False):
  """Противоположность функции скрытия файла"""
  return hide_file(path, recursive=recursive, unhide=not hide)


class LnkFile:
  """Открыть/создать файл `.lnk`"""

  def __init__(self, path: PATH_TYPES, readonly: bool = True):
    path = path2str(path, True).replace("/", "\\")
    if readonly:
      if not os.path.isfile(path):
        raise ms.file.NotAFileError(path)
    self.obj: win32com.client.CDispatch = shell["WScript"].CreateShortcut(path)
    self.readonly = readonly

  def __getitem__(self, k):
    result = getattr(self, k, None)
    if result == "":
      return None
    return result

  def __setitem__(self, k, v):
    if self.readonly:
      raise RuntimeError("Read-only shortcut")
    if v is None:
      v = ""
    setattr(self.obj, k, v)

  def save(self):
    """Сохранить изменения"""
    if self.readonly:
      raise RuntimeError("Read-only shortcut")
    self.obj.Save()

  @property
  def args_str(self) -> None | str:
    return self["Arguments"]

  @args_str.setter
  def args_str(self, v):
    self["Arguments"] = v

  @property
  def args(self) -> None | list[str]:
    line = self.args_str
    if line is None:
      return None
    import shlex
    return shlex.split(line, posix=False)

  @args.setter
  def args(self, v):
    if v is None:
      self["Arguments"] = None
      return
    import shlex
    self["Arguments"] = shlex.join(v)

  @property
  def description(self) -> None | str:
    return self["Description"]

  @description.setter
  def description(self, v):
    self["Description"] = v

  @property
  def hot_key(self) -> None | str:
    return self["Hotkey"]

  @hot_key.setter
  def hot_key(self, v):
    self["Hotkey"] = v

  @property
  def icon_loc(self) -> None | str:
    return self["IconLocation"]

  @icon_loc.setter
  def icon_loc(self, v):
    self["IconLocation"] = v

  @property
  def path(self) -> str:
    return self["FullName"]

  @property
  def target_path(self) -> None | str:
    return self["TargetPath"]

  @target_path.setter
  def target_path(self, v):
    self["TargetPath"] = path2str(v, True).replace("/", "\\")

  @property
  def window_style(self) -> int:
    return self["WindowStyle"]

  @window_style.setter
  def window_style(self, v):
    self["WindowStyle"] = v

  @property
  def working_dir(self) -> None | str:
    return self["WorkingDirectory"]

  @working_dir.setter
  def working_dir(self, v):
    self["WorkingDirectory"] = v



import argparse
import os
import sys
from MainShortcuts2 import ms
HASH_SUFFIX = ".MS2_hash"
HASH_TYPES = ["blake2b", "blake2s", "md5", "sha1", "sha224", "sha256", "sha384", "sha3_224", "sha3_256", "sha3_384", "sha3_512", "sha512"]


def import_example():
  argp = argparse.ArgumentParser("ms2-import_example", description="код для импорта MainShortcuts2")
  argp.parse_args()
  print("from MainShortcuts2 import ms")


def nano_json(args: argparse.Namespace = None):
  ms.utils.check_programs("nano")
  import subprocess
  if args is None:
    argp = argparse.ArgumentParser("nano-json", description="форматирование JSON файлов и редактирование в GNU NANO")
    argp.add_argument("files", nargs="+", help="пути к файлам JSON")
    argp.add_argument("--nano-help", action="store_true", help="показать помощь nano")
    argp.add_argument("-e", "--encoding", default="utf-8", help="кодировка файлов")
    argp.add_argument("-f", "--rcfile", help="использовать только этот файл для настройки nano")
    argp.add_argument("-i", "--indent", default=2, type=int, help="кол-во пробелов для отступа")
    argp.add_argument("-m", "--mode", choices=ms.json.MODES_ALL, default="p", help="режим сохранения редактирования")
    argp.add_argument("-s", "--sort", action="store_true", help="сортировать ключи словаря")
    argp.add_argument("-u", "--no-escape", action="store_false", help="не использовать Unicode Escape")
    args = argp.parse_args()
  if args.nano_help:
    return subprocess.call(["nano", "--help"])
  write_kw = {}
  write_kw["encoding"] = args.encoding
  write_kw["ensure_ascii"] = False
  write_kw["indent"] = args.indent
  write_kw["mode"] = "print"
  write_kw["sort_keys"] = args.sort
  nano_args = ["nano"]
  if args.rcfile:
    nano_args += ["--rcfile", args.rcfile]
  nano_args += args.files
  for i in args.files:
    try:
      data = ms.json.read(i, encoding=args.encoding)
      ms.json.write(i, data, **write_kw)
    except Exception as err:
      print(err, file=sys.stderr)
  subprocess.call(nano_args)
  write_kw["ensure_ascii"] = args.no_escape
  write_kw["indent"] = None if args.mode in ms.json.MODES_COMPRESS else args.indent
  write_kw["mode"] = args.mode
  for i in args.files:
    try:
      data = ms.json.read(i, encoding=args.encoding)
      ms.json.write(i, data, **write_kw)
    except Exception as err:
      print(err, file=sys.stderr)


def _check_nginx() -> int:
  import subprocess
  with subprocess.Popen(["nginx", "-t"], stderr=subprocess.PIPE) as p:
    code = p.wait()
    if code != 0:
      sys.stderr.buffer.write(p.stderr.read())
  return code


def nginx_reload():
  ms.utils.check_programs("nginx")
  import subprocess
  argp = argparse.ArgumentParser("nginx-reload", description="проверка конфига и перезагрузка Nginx")
  argp.parse_args()
  code = _check_nginx()
  if code != 0:
    sys.exit(code)
  sys.exit(subprocess.call(["nginx", "-s", "reload"]))


def nginx_restart():
  ms.utils.check_programs("nginx", "systemctl")
  import subprocess
  argp = argparse.ArgumentParser("nginx-reload", description="проверка конфига и перезапуск Nginx через systemctl")
  argp.parse_args()
  code = _check_nginx()
  if code != 0:
    sys.exit(code)
  sys.exit(subprocess.call(["systemctl", "restart", "nginx"]))


def hash_gen(args: argparse.Namespace = None):
  import hashlib
  import os
  import shlex
  if args is None:
    argp = argparse.ArgumentParser("ms2-hash_gen", description="создание контрольной суммы для файла")
    argp.add_argument("files", nargs="+", help="пути к файлам")
    argp.add_argument("-b", "--bar", action="store_true", help="показывать прогрессбар (нужен модуль progressbar2)")
    argp.add_argument("-f", "--force", action="store_true", help="перезаписывать существующие хеши")
    argp.add_argument("-t", "--type", choices=HASH_TYPES, default="sha512", help="тип контрольной суммы")
    args = argp.parse_args()
  pbar, pbar_w, progressbar = None, None, None
  if args.bar:
    import progressbar
    pbar_w = [progressbar.Percentage(), progressbar.GranularBar(left="(", right=")"), progressbar.FileTransferSpeed(), " ", progressbar.ETA(format="%(eta)8s", format_finished="%(elapsed)8s", format_na="     N/A", format_not_started="--:--:--", format_zero="00:00:00")]
  completed = []
  data = {}
  data["file"] = {}
  data["format"] = "MainShortcuts2_hash_v1"
  data["hash"] = {}
  data["hash"]["type"] = args.type
  Hash = getattr(hashlib, args.type)
  for file in args.files:
    while file.lower().endswith(HASH_SUFFIX.lower()):
      file = file[:0 - len(HASH_SUFFIX)]
    if file in completed:
      continue
    if os.path.isdir(file):
      print("Пропуск файла " + shlex.quote(file) + ": это папка", file=sys.stderr)
      completed.append(file)
      continue
    if os.path.isfile(file + HASH_SUFFIX):
      if not args.force:
        print("Пропуск файла " + shlex.quote(file) + ": хеш существует", file=sys.stderr)
        continue
    data["file"]["size"] = os.path.getsize(file)
    with open(file, "rb") as f:
      if args.bar:
        pbar = progressbar.ProgressBar(max_error=False, max_value=data["file"]["size"], widgets=[ms.path.Path(file).full_name] + pbar_w)
        c = 0
        pbar.start()
      hash = Hash()
      for chunk in f:
        hash.update(chunk)
        if args.bar:
          c += len(chunk)
          pbar.update(c)
    if args.bar:
      pbar.finish()
    data["hash"]["hex"] = hash.hexdigest()
    ms.json.write(file + HASH_SUFFIX, data)
    completed.append(file)


def hash_check(args: argparse.Namespace = None):
  import hashlib
  import shlex
  if args is None:
    argp = argparse.ArgumentParser("ms2-hash_check", description="проверка размера и контрольной суммы файла")
    argp.add_argument("files", nargs="+", help="пути к файлам")
    argp.add_argument("-b", "--bar", action="store_true", help="показывать прогрессбар (нужен модуль progressbar2)")
    args = argp.parse_args()
  pbar, pbar_w, progressbar = None, None, None
  if args.bar:
    import progressbar
    pbar_w = [progressbar.Percentage(), progressbar.GranularBar(left="(", right=")"), progressbar.FileTransferSpeed(), " ", progressbar.ETA(format="%(eta)8s", format_finished="%(elapsed)8s", format_na="     N/A", format_not_started="--:--:--", format_zero="00:00:00")]
  completed = []
  for file in args.files:
    while file.lower().endswith(HASH_SUFFIX.lower()):
      file = file[:0 - len(HASH_SUFFIX)]
    if file in completed:
      continue
    if os.path.isdir(file):
      print("Пропуск файла " + shlex.quote(file) + ": это папка", file=sys.stderr)
      completed.append(file)
      continue
    if not os.path.exists(file + HASH_SUFFIX):
      print("Ошибка: не найден файл " + shlex.quote(file + HASH_SUFFIX), file=sys.stderr)
      continue
    data = ms.json.read(file + HASH_SUFFIX)
    if data["file"]["size"] != os.path.getsize(file):
      print("Ошибка: размер файла " + shlex.quote(file) + " не совпадает", file=sys.stderr)
      continue
    if args.bar:
      pbar = progressbar.ProgressBar(max_error=False, max_value=data["file"]["size"], widgets=[ms.path.Path(file).full_name] + pbar_w)
      c = 0
      pbar.start()
    hash = getattr(hashlib, data["hash"]["type"])()
    with open(file, "rb") as f:
      for chunk in f:
        hash.update(chunk)
        if args.bar:
          c += len(chunk)
          pbar.update(c)
    if args.bar:
      pbar.finish()
    if data["hash"]["hex"] == hash.hexdigest():
      print("Успех: файл " + shlex.quote(file) + " не изменён")
    else:
      print("Ошибка: файл " + shlex.quote(file) + " изменён")
    completed.append(file)


def ln(args=None):
  if args is None:
    argp = argparse.ArgumentParser("ms2-ln", description="создание символьной ссылки с абсолютным путём к файлу")
    argp.add_argument("-f", "--force", action="store_true", help="принудительное создание ссылки")
    argp.add_argument("file", help="путь к файлу, на который будет указывать ссылка")
    argp.add_argument("link", help="путь к ссылке для создания")
    args = argp.parse_args()
  if args.force:
    ms.path.delete(args.link)
  file = os.path.realpath(args.file)
  os.symlink(file, args.link, os.path.isdir(file))



"""Работа с JSON"""
import atexit
import builtins
import json
from .core import ms
from .path import PATH_TYPES
from typing import *
try:
  import json5
except Exception:
  json5 = None
JSON_TYPES = Union[bool, dict, float, int, list, None, str]
MODES_COMPRESS = ("c", "compress", "min", "zip")
MODES_MP = ("mp", "mp_tg", "mainplay", "mainplay_tg")
MODES_PRETTY = ("p", "pretty", "max", "print")
MODES_ALL = MODES_COMPRESS + MODES_MP + MODES_PRETTY


def decode(text: str, *, like_json5: bool = False, **kw) -> JSON_TYPES:
  """Прочитать текстовый JSON"""
  kw["s"] = text
  if like_json5:
    if not json5 is None:
      return json5.loads(**kw)
  return json.loads(**kw)


def encode(data: JSON_TYPES, mode: str = "c", **kw):
  """Создать текстовый JSON"""
  kw["obj"] = data
  mode = mode.lower()
  if "force" in kw:
    del kw["force"]
    import warnings
    warnings.warn("The argument 'force' is no longer used", DeprecationWarning)
  if "sort" in kw:
    kw["sort_keys"] = kw.pop("sort")
  if mode in MODES_COMPRESS:  # Сжатый
    kw.setdefault("indent", None)
    kw.setdefault("separators", [",", ":"])
  if mode in MODES_PRETTY:  # Развёрнутый
    kw.setdefault("indent", 2)
    kw.setdefault("separators", [",", ": "])
    if mode == "print":
      kw.setdefault("ensure_ascii", False)
  if mode in MODES_MP:  # Стиль MainPlay TG
    kw.setdefault("indent", 2)
    kw.setdefault("separators", [",", ":"])
  return json.dumps(**kw)


def print(data: JSON_TYPES, **kw):
  """Напечатать данные в виде текстового JSON"""
  pr_kw = {}
  for i in ["end", "file", "flush", "sep"]:
    if i in kw:
      pr_kw[i] = kw.pop(i)
  kw["data"] = data
  kw.setdefault("mode", "print")
  builtins.print(encode(**kw), **pr_kw)


def read(path: PATH_TYPES, **kw) -> JSON_TYPES:
  """Прочитать JSON файл"""
  f_kw = {}
  if "encoding" in kw:
    f_kw["encoding"] = kw.pop("encoding")
  f_kw["path"] = path
  kw["text"] = ms.file.read(**f_kw)
  return decode(**kw)


def rebuild(text: str, **kw) -> str:
  """Перестроить текстовый JSON"""
  de_kw = {}
  if "like_json5" in kw:
    de_kw["like_json5"] = kw.pop("like_json5")
  de_kw["text"] = text
  kw["data"] = decode(**de_kw)
  return encode(**kw)


def rewrite(path: PATH_TYPES, **kw) -> int:
  """Перестроить JSON файл"""
  de_kw = {}
  if "encoding" in kw:
    de_kw["encoding"] = kw["encoding"]
  if "like_json5" in kw:
    de_kw["like_json5"] = kw.pop("like_json5")
  de_kw["path"] = path
  kw["path"] = path
  kw["data"] = read(**de_kw)
  return write(**kw)


def write(path: PATH_TYPES, data: JSON_TYPES, **kw) -> int:
  """Прочитать JSON файл"""
  f_kw = {}
  kw["data"] = data
  for i in ["encoding", "force", "use_tmp_file"]:
    if i in kw:
      f_kw[i] = kw.pop(i)
  f_kw["path"] = path
  f_kw["data"] = encode(**kw)
  return ms.file.write(**f_kw)


class JsonFile:
  def __init__(self, path: str, autoload: bool = True, save_at_exit: bool = False, **kw):
    atexit.register(self.__exit__)
    self._data = None
    self._loaded = False
    self._path = None
    self.load_kw = {}
    self.save_kw = kw
    self.path = ms.path.Path(path)
    self.save_at_exit = save_at_exit
    if "like_json5" in self.save_kw:
      self.load_kw["like_json5"] = self.save_kw.pop("like_json5")
    if autoload:
      self.load()

  def __contains__(self, k) -> bool:
    return k in self.data

  def __delitem__(self, k):
    del self.data[k]

  def __enter__(self):
    return self

  def __exit__(self, *a):
    if self.save_at_exit:
      self.save()

  def __getitem__(self, k):
    return self.data[k]

  def __setitem__(self, k, v):
    self.data[k] = v

  @property
  def path(self) -> ms.path.Path:
    return self._path

  @path.setter
  def path(self, v):
    self._path = ms.path.Path(v)
    self.load_kw["path"] = self._path.path
    self.save_kw["path"] = self._path.path

  @property
  def data(self):
    if not self._loaded:
      self.load()
    return self._data

  @data.setter
  def data(self, v):
    self._data = v
    self._loaded = True

  def load(self):
    self._data = read(**self.load_kw)
    self._loaded = True

  def save(self):
    self.save_kw["data"] = self.data
    return write(**self.save_kw)



"""Упрощение и сокращение вашего кода благодаря этой библиотеке
Разработчик: MainPlay TG
https://t.me/MainPlayCh"""
__all__ = ["ms", "NoLogger"]
__scripts__ = []
from ._module_info import version as __version__
# 2.0.0
from .core import ms, NoLogger



"""Работа со строками"""
from .core import ms
from typing import *


def list2str(a: Union[Iterable]) -> list:
  """Преобразовать каждый элемент списка в строку"""
  b = []
  for i in a:
    b.append(str(i))
  return b


def dict2str(a: dict) -> dict:
  """Преобразовать каждое значение словаря в строку"""
  b = {}
  for key, value in a.items():
    b[key] = str(value)
  return b


class _Replace:
  """Функции для замены текста"""

  def __init__(self):
    pass

  def multi(self, text: str, d: dict = None, **kw) -> str:
    """Мульти-замена {"что заменить":"чем заменить"}"""
    if not d is None:
      kw.update(d)
    t = str(text)
    for k, v in d.items():
      t = t.replace(k, str(v))
    return t

  def all(self, text: str, fr: str, to: str) -> str:
    """Замена пока заменяемый текст не исчезнет"""
    t = str(text)
    a = str(fr)
    b = str(to)
    if a in b:
      raise RecursionError('%r is contained in %r, this causes an infinite loop' % (a, b))
    while a in t:
      t = t.replace(a, b)
    return t


replace = _Replace()



"""Работа со списками"""
import re
from .core import ms
from typing import *


def filter(a: list, whitelist: list = None, blacklist: list = [], regex: str = False, begin: str = None, end: str = None):
  """Фильтровать список"""
  a = list(a)
  if whitelist == None:
    whitelist = a
  if type(whitelist) == str:
    whitelist = [whitelist]
  if type(blacklist) == str:
    blacklist = [blacklist]
  b = []
  for i in a:
    add = True
    if begin != None:
      if str(i).startswith(str(begin)):
        add = True
      else:
        add = False
    if end != None and add:
      if str(i).endswith(str(end)):
        add = True
      else:
        add = False
    if regex and add:
      reW = False
      for i2 in whitelist:
        if re.match(str(i2), str(i)) != None:
          reW = True
          break
      reB = False
      for i2 in blacklist:
        if re.match(str(i2), str(i)) != None:
          reB = True
          break
      if reW and not reB:
        add = True
      else:
        add = False
    if add and not regex:
      if (i in whitelist) and (not i in blacklist):
        add = True
      else:
        add = False
    if add:
      b.append(i)
  return b


def rm_duplicates(a: list, trim: bool = False, case: bool = False, func=lambda i: i):
  """Удалить дублирующиеся элементы"""
  b = []
  trim = str(trim).lower()
  case = str(case).lower()
  for i in a:
    if trim in ["true", "lr", "rl", "all"]:
      i = i.strip()
    elif trim in ["l", "left"]:
      i = i.lstrip()
    elif trim in ["r", "right"]:
      i = i.rstrip()
    if case in ["l", "lower", "low"]:
      i = i.lower()
    elif case in ["u", "upper", "up"]:
      i = i.upper()
    elif case in ["capitalize", "cap"]:
      i = i.capitalize()
    i = func(i)
    if not i in b:
      b.append(i)
  return b



import re
from .core import ms
from .path import Path
from typing import Any, Union


def _check_count(data):
  if len(data) == 0:
    raise ValueError("The list is empty")


class MultiLang:
  FORMAT = 1

  def __init__(self, default_lang: Union[dict, Path, str]):
    self.cache: dict[tuple[str, str, str], str] = {}
    self.cache_builders: dict = {}
    self.files: dict[str, Path] = {}
    self.langs: dict[Union[None, str], dict[str, dict[str, Any]]] = {}

    @self.add_cache_builder("lines")
    def _(text: list[str]) -> str:
      return "\n".join(text)

    @self.add_cache_builder("normal")
    def _(text: str) -> str:
      assert type(text) == str
      return text
    self.add_lang(None, default_lang, load=True)

  def add_cache_builder(self, name: str):
    def deco(func):
      self.cache_builders[name] = func
      return func
    return deco

  def add_langs(self, langs: dict[str, Union[dict, Path, str]], check_count: bool = True, load: bool = True):
    """Добавить языки (`dict`) или пути к языковым файлам (`str`, `ms.path.Path`)"""
    if check_count:
      _check_count(langs)
    for k, v in langs.items():
      self.add_lang(k, v, load=load)

  def add_lang(self, lang_name: str, lang: Union[dict, Path, str], load: bool = True):
    """Добавить один язык из словаря или пути к файлу"""
    if type(lang) == dict:
      for cat_name, cat in lang.items():
        assert type(cat_name) == str
        assert type(cat) == dict
        for text_name, text in cat.items():
          assert type(text_name) == str
          if type(text) == dict:
            if not "allow_cache" in text:
              text["allow_cache"] = True
            if not "type" in text:
              text["type"] = "normal"
      self.langs[lang_name] = lang
      return
    if type(lang) == str:
      lang = Path(lang)
    if type(lang) == Path:
      self.files[lang_name] = lang
      if load:
        self.load(lang_name)
      return

  def load(self, *names: Union[None, str], check_count: bool = True):
    """Загрузить языковые файлы"""
    if check_count:
      _check_count(names)
    for i in names:
      path = self.files[i].path
      data = ms.json.read(path)
      self.add_lang(i, data["texts"])

  def save(self, *names: Union[None, str], check_count: bool = True, **kw) -> int:
    """Сохранить языковые файлы"""
    if check_count:
      _check_count(names)
    data = {"format": "MainShortcuts2.advanced.MultiLang/%i" % self.FORMAT}
    sum = 0
    for i in names:
      data["texts"] = self.langs[i]
      kw["data"] = data
      kw["path"] = self.files[i].path
      sum += ms.json.write(**kw)
    return sum

  def build_cache(self, lang_name: Union[None, str], cat_name: str, text_name: str):
    cache_name = (lang_name, cat_name, text_name)
    if cache_name in self.cache:
      return
    text = self.langs[lang_name][cat_name][text_name]
    if type(text) == str:
      self.cache[cache_name] = text
      return
    builder = self.cache_builders[text["type"]]
    result = builder(text)
    assert type(result) == str
    self.cache[cache_name] = result

  def get_text(self, cat: str, name: str, values=None, *, lang: str = None) -> str:
    if not lang in self.langs:
      self.load(lang)
    raw = self.langs[cat][name]
    if type(raw) == str:
      text = raw
    else:
      if raw.get("allow_cache") == False:
        text = self.cache_builders[raw["type"]](raw)
      else:
        self.build_cache(lang, cat, raw)
        text = self.cache[lang, cat, name]
    if values is None:
      return text
    return text % values


class PermissionSystem:
  FORMAT = 1

  def __init__(self, path: str):
    self.path = Path(path)
    self.load()

  def load(self):
    data = ms.json.read(self.path.path)
    self.default_groups = data["default_groups"] if "default_groups" in data else []
    self.default_perms = data["default_perms"] if "default_perms" in data else {}
    self.groups = {}
    self.users = {}
    for i in ["all", "group", "user"]:
      if not i in self.default_perms:
        self.default_perms[i] = {}
    if "groups" in data:
      for name, item in data["groups"].items():
        if not "admin" in item:
          item["admin"] = False
        if not "perms" in item:
          item["perms"] = self.default_perms["all"].copy()
          item["perms"].update(self.default_perms["groups"])
        if not "priority" in item:
          item["priority"] = 0
        self.groups[name] = item
    if "users" in data:
      for name, item in data["users"].items():
        if not "admin" in item:
          item["admin"] = False
        if not "groups" in item:
          item["groups"] = self.default_groups
        if not "perms" in item:
          item["perms"] = self.default_perms["all"].copy()
          item["perms"].update(self.default_perms["users"])
        self.groups[name] = item

  def save(self, **kw):
    kw["path"] = self.path.path
    kw["data"] = {"format": "MainShortcuts2.advanced.PermissionSystem/%i" % self.FORMAT}
    kw["data"]["default_groups"] = self.default_groups
    kw["data"]["default_perms"] = self.default_perms
    kw["data"]["groups"] = self.groups
    kw["data"]["users"] = self.users
    return ms.json.write(**kw)

  def _i_compare(self, perms: dict[str, bool], perm: str) -> bool:
    for k, v in perms:
      if re.match(k, perm):
        return v

  def _verify(self, username: str, permname: str) -> bool:
    if not username in self.users:
      return False
    user = self.users[username]
    if user["admin"]:
      return True
    if permname in user["perms"]:
      return user["perms"][permname]
    group = None
    for i in user["groups"]:
      if i in self.groups:
        if self.groups[i]["admin"]:
          return True
        if permname in self.groups[i]["perms"]:
          if group is None:
            group = self.groups[i]
            continue
          if self.groups[i]["priority"] > group["priority"]:
            group = self.groups[i]
            continue
    if group is None:
      return False
    return group["perms"][permname]

  def verify(self, username: str, permname: str, raise_error: bool = True) -> bool:
    """Проверить есть ли у пользователя данное право. Если пользователь является админом или одна из его групп даёт разрешения админа, права есть. Если права не указаны в пользователе, используются права группы с наивысшим приоритетом, в которой указано это право. Если ни у одной группы нет права, значит права отсутствуют"""
    result = self._verify(username, permname)
    if raise_error:
      if not result:
        raise ms.types.AccessDeniedError(username, permname)
    return result

  def add_group(self, name: str, perms: dict[str, bool] = None, *, add_defaults_perms: bool = True, admin: bool = False, priority: int = 0):
    group = {
        "admin": admin,
        "perms": {} if perms is None else perms.copy(),
        "priority": priority,
    }
    if add_defaults_perms:
      for i in self.default_perms["all"]:
        if not i in group["perms"]:
          group["perms"][i] = self.default_perms["all"][i]
      for i in self.default_perms["group"]:
        if not i in group["perms"]:
          group["perms"][i] = self.default_perms["group"][i]
    self.groups[name] = group

  def add_user(self, name: str, perms: dict[str, bool] = None, *, add_defaults_perms: bool = True, admin: bool = False, groups: list[str] = None):
    user = {
        "admin": admin,
        "groups": [] if groups is None else groups,
        "perms": {} if perms is None else perms.copy(),
    }
    if add_defaults_perms:
      for i in self.default_perms["all"]:
        if not i in user["perms"]:
          user["perms"][i] = self.default_perms["all"][i]
      for i in self.default_perms["user"]:
        if not i in user["perms"]:
          user["perms"][i] = self.default_perms["user"][i]
    self.users[name] = user

  def edit_group(self, name: str, *, admin: bool = None, perms: dict[str, Union[None, bool]] = None, priority: int = None):
    group = self.groups[name]
    if not admin is None:
      group["admin"] = bool(admin)
    if not priority is None:
      group["priority"] = int(priority)
    if not perms is None:
      for k, v in perms.items():
        if v is None:
          if k in group["perms"]:
            del group["perms"][k]
          else:
            group["perms"][k] = v

  def edit_user(self, name: str, *, admin: bool = None, groups: list[str] = None, perms: dict[str, Union[None, bool]] = None):
    user = self.users[name]
    if not admin is None:
      user["admin"] = bool(admin)
    if not groups is None:
      for i in groups:
        assert type(i) == str
      user["groups"] = groups
    if not perms is None:
      for k, v in perms.items():
        if v is None:
          if k in user["perms"]:
            del user["perms"][k]
          else:
            user["perms"][k] = v


class DictScriptAction:
  def __init__(self, name: str, args: list = None, kwargs: dict = None, save_to: str = None, comment: str = None):
    self.__getitem__ = self.__getattr__
    self.args: list = args
    self.comment: str = comment
    self.kwargs: dict = kwargs
    self.name: str = name
    self.save_to: str = save_to

  @classmethod
  def from_dict(cls, data: dict):
    kw = {}
    for k in ["args", "comment", "kwargs", "name", "save_to"]:
      kw[k] = data.get(k)
    return cls(**kw)

  def to_dict(self) -> dict:
    result = {}
    for k in ["args", "comment", "kwargs", "name", "save_to"]:
      v = getattr(self, k)
      if not v is None:
        result[k] = v
    return result


class DictScriptVariable:
  def __init__(self, name: str):
    self.name: str = name


@ms.any2json.reg_decoder(DictScriptVariable)
def _(obj):
  return DictScriptVariable(obj)


@ms.any2json.reg_encoder(DictScriptVariable)
def _(obj: DictScriptVariable):
  if isinstance(obj, DictScriptVariable):
    return obj.name


class DictScriptRunner:
  VERSION = 1

  def __init__(self, functions: dict = None, add_default_functions: bool = True):
    self.globals = {}
    if add_default_functions:
      import time
      for i in [bool, bytes, dict, float, int, list, str, tuple]:
        self.reg_class()(i)
      self.reg_function("sleep")(time.sleep)
      self.reg_function("sum")(sum)
      self.reg_function("time.sleep")(time.sleep)
      self.reg_function("time.time")(time.time)
    if not functions is None:
      for name, func in functions.items():
        self.reg_function(name)(func)

  def reg_class(self, name: str = None, overwrite: bool = False):
    if not name is None:
      import warnings
      warnings.warn("The argument 'name' temporarily does not work", FutureWarning)

    def deco(cls: type) -> type:
      name = cls.__module__ + "." + cls.__name__
      if callable(cls):
        self.reg_function(name, overwrite=overwrite)(cls)
      for k in dir(cls):
        v = getattr(cls, k)
        if callable(v):
          self.reg_function(name + "." + k, overwrite=overwrite)
      return cls
    return deco

  def reg_function(self, name: str, overwrite: bool = False):
    if not overwrite:
      if name in self.globals:
        raise ValueError("The %r function already exists. Perhaps you wanted to overwrite it?" % name)

    def deco(func):
      if not callable(func):
        raise TypeError("Function %r is not callable" % name)
      self.globals[name] = func
      return func
    return deco

  def run_script(self, script: list[dict]) -> dict:
    """Выполнить скрипт (список действий)"""
    for act in script:
      if not act["name"] in self.globals:
        raise KeyError("Function %r not exists" % act["name"])
    locals = self.globals.copy()
    for act in script:
      if act["name"] == "exit":
        break
      args = act["args"] if "args" in act else []
      func = locals[act["name"]]
      kwargs = act["kwargs"] if "kwargs" in act else {}
      result = func(*args, **kwargs)
      if "save_to" in act:
        locals[act["save_to"]] = result
    return locals


class CodeModule:
  """Импорт модуля из исходного кода. Могут быть баги"""

  def __init__(self, source: str, globals: dict = None, locals: dict = None):
    if globals is None:
      globals = {}
    if locals is None:
      locals = {}
    args = (source, globals, locals)
    exec(*args)
    self.__dict__["source"] = args[0]
    self.__dict__["globals"] = args[1]
    self.__dict__["globals"].update(args[2])

  def __delattr__(self, k):
    del self.__dict__["globals"][k]

  def __dir__(self) -> list[str]:
    return list(self.__dict__["globals"])

  def __getattr__(self, k):
    return self.__dict__["globals"][k]

  def __hasattr__(self, k) -> bool:
    return k in self.__dict__["globals"]

  def __setattr__(self, k, v):
    self.__dict__["globals"][k] = v



import builtins
from .core import ms
decoders = {}
encoders = {}


def reg_decoder(type: str):
  if builtins.type(type) != str:
    type = type.__module__ + "." + type.__name__

  def deco(func):
    decoders[type] = func
    return func
  return deco


def reg_encoder(type: str):
  if builtins.type(type) != str:
    type = type.__module__ + "." + type.__name__

  def deco(func):
    encoders[type] = func
    return func
  return deco


@reg_decoder(type(None))
def _(obj):
  return None


@reg_decoder(bool)
def _(obj):
  return obj


@reg_decoder(bytes)
def _(obj):
  import base64
  if obj[0] == "base64":
    return base64.b64decode(obj[1].encode("utf-8"))


@reg_decoder(dict)
def _(obj):
  result = {}
  for k, v in obj:
    result[ms.any2json._decode_obj(k)] = ms.any2json._decode_obj(v)
  return result


@reg_decoder(float)
def _(obj):
  return obj


@reg_decoder(int)
def _(obj):
  return obj


@reg_decoder(list)
def _(obj):
  return [ms.any2json._decode_obj(i) for i in obj]


@reg_decoder(str)
def _(obj):
  return obj


@reg_decoder(tuple)
def _(obj):
  return tuple([ms.any2json._decode_obj(i) for i in obj])


@reg_encoder(type(None))
def _(obj):
  if obj is None:
    return 0


@reg_encoder(bool)
def _(obj):
  if isinstance(obj, bool):
    return bool(obj)


@reg_encoder(bytes)
def _(obj):
  if isinstance(obj, bytes):
    from base64 import b64encode
    result = [
        "base64",
        b64encode(obj).decode("utf-8"),
    ]
    return result


@reg_encoder(dict)
def _(obj):
  if isinstance(obj, dict):
    result = []
    for k, v in obj.items():
      result.append([ms.any2json._encode_obj(k), ms.any2json._encode_obj(v)])
    return result


@reg_encoder(float)
def _(obj):
  if isinstance(obj, float):
    return float(obj)


@reg_encoder(int)
def _(obj):
  if isinstance(obj, int):
    return int(obj)


@reg_encoder(list)
def _(obj):
  if isinstance(obj, list):
    return [ms.any2json._encode_obj(i) for i in obj]


@reg_encoder(str)
def _(obj):
  if isinstance(obj, str):
    return str(obj)


@reg_encoder(tuple)
def _(obj):
  if isinstance(obj, tuple):
    return [ms.any2json._encode_obj(i) for i in obj]



import os
import requests
from .base import BaseClient
from io import IOBase
from MainShortcuts2 import ms
UPLOAD_HEADERS = {"Content-Type": "application/octet-stream"}


class _ReadBinaryIO(IOBase):
  def __init__(self, resp: requests.Response):
    IOBase.__init__(self)
    self.mode: str = "rb"
    self.resp = resp

  def __exit__(self, *a):
    self.close()

  def close(self):
    self.resp.close()

  def read(self, size: int = None) -> bytes:
    return self.resp.raw.read(size)

  def readable(self) -> bool:
    return True

  def seekable(self) -> bool:
    return False

  def writable(self) -> bool:
    return False


class WebDAVClient(BaseClient):
  """Простой WebDAV клиент"""

  def __init__(self, root: str, **kw):
    self._init(**kw)
    self._cwd: str = "/"
    self._url = (root[:-1] if root.endswith("/") else root) + "{method}"

  def _auth(self):
    pass  # Без авторизации

  def _fspath(self, path) -> str:
    if isinstance(path, bytes):
      path = path.decode("utf-8")
    return os.fspath(path)

  @property
  def cwd(self) -> str:
    """Рабочая папка на сервере"""
    return self._cwd

  @cwd.setter
  def cwd(self, value: str):
    path = self._fspath(value)
    if not path.startswith("/"):
      path = self.cwd + "/" + path
    self._cwd = os.path.normpath(path[:-1] if path.endswith("/") else path)

  def request(self, method: str, path: str, **kw) -> requests.Response:
    self._auth()
    path = self._fspath(path)
    if not path.startswith("/"):
      path = self.cwd + path
    return self._request(method, path, **kw)

  def open(self, path: str, mode: str = "rb") -> _ReadBinaryIO:
    """Открыть файл сервера (read-only)"""
    if not mode in ["rb", "br"]:
      raise ValueError("Only the 'rb' mode is supported")
    return _ReadBinaryIO(self.request("GET", path, stream=True))

  def download(self, path: str, filepath: str, **kw) -> int:
    """Скачать файл с сервера"""
    return ms.utils.download_file(self.request("GET", path, stream=True), filepath)

  def upload(self, filepath: str, path: str):
    """Загрузить локальный файл на сервер"""
    with open(filepath, "rb") as f:
      self.request("PUT", path, data=f, headers=UPLOAD_HEADERS)

  def mkdir(self, path: str):
    """Создать пустую папку на сервере"""
    if not path.endswith("/"):
      path += "/"
    self.request("MKCOL", path)

  def delete(self, path: str):
    """Удалить файл/папку на сервере"""
    self.request("DELETE", path)



"""Различные объекты и исключения"""
from typing import Union


class Base:
  def __init__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs
    self.type = type(self)


class UserError(Exception):
  """Ошибка, которую допустил пользователь. Например неправильно указал входные данные"""
  pass


class AccessDeniedError(UserError):
  """Ошибка доступа"""


class Empty(Base):
  """Пустота (не равно `None`)"""

  def __bool__(self):
    return False


class Infinity(Base):
  """Бесконечное число"""

  def __bool__(self):
    return True


class NotAFileError(Exception):
  """Ошибка 'это не файл'"""


class NotANumber(Base):
  """Не число"""


class NotFound(Base):
  """Не найдено"""

  def __bool__(self):
    return False


class NotFoundError(Exception):
  """Ошибка 'не найдено'"""


class Action:
  """Запланированный запуск функции"""

  def __init__(self, func, *args, **kwargs):
    self._closed = False
    self._completed = False
    self._error = None
    self._launched = False
    self._result = None
    self.args: tuple = args
    self.func = func
    self.kwargs: dict = kwargs

  def __enter__(self):
    return self

  def __exit__(self, a, b, c):
    self.close()

  def _check(self, launched: bool = None, completed: bool = None, closed: bool = None):
    if not launched is None:
      if launched:
        if not self.launched:
          raise RuntimeError("The action has not yet been launched")
      else:
        if self.launched:
          raise RuntimeError("The action has already been launched")
    if not completed is None:
      if completed:
        if not self.completed:
          raise RuntimeError("The action has not yet been completed")
      else:
        if self.completed:
          raise RuntimeError("The action has already been completed")
    if not closed is None:
      if closed:
        if not self.closed:
          raise RuntimeError("The action has not yet been closed")
      else:
        if self.closed:
          raise RuntimeError("The action has already been closed")

  def close(self):
    self._closed = True
    self._error = None
    self._result = None
    self.args = None
    self.func = None
    self.kwargs = None

  @property
  def launched(self) -> bool:
    """Была ли запущена функция"""
    return self._launched

  @property
  def completed(self) -> bool:
    """Завершена ли функция"""
    return self._completed

  @property
  def exception(self) -> None | Exception:
    """Ошибка, возникшая при выполнении функции"""
    self._check(launched=True, completed=True, closed=False)
    return self._error

  @property
  def result(self):
    """Результат выполнения функции"""
    self._check(launched=True, completed=True, closed=False)
    if not self._error is None:
      raise self._error  # type: ignore
    return self._result

  @property
  def closed(self) -> bool:
    return self._closed

  def run(self):
    """Запустить функцию"""
    self._check(launched=False, completed=False, closed=False)
    self._launched = True
    try:
      self._result = self.func(*self.args, **self.kwargs)  # type: ignore
    except Exception as err:
      self._error = err
    self._completed = True


class OnlyOneInstanceError(BaseException):
  """Ошибка для `.utils.OnlyOneInstance`"""

  def __init__(self, *args):
    if len(args) == 0:
      args = ("Another instance is already running",)
    BaseException.__init__(self, *args)


class DotDict:
  def __init__(self, data: dict = None):
    self._data = {}
    for k in data:
      self[k] = data[k]

  def __delattr__(self, k: str):
    if k.startswith("_"):
      return object.__delattr__(self, k)
    del self.data[k]

  def __getattr__(self, k: str):
    if k.startswith("_"):
      return object.__getattr__(self, k)
    return self[k]

  def __setattr__(self, k: str, v):
    if k.startswith("_"):
      return object.__setattr__(self, k, v)
    self[k] = v

  def __contains__(self, k: str):
    return k in self._data

  def __delitem__(self, k: str):
    del self._data[k]

  def __getitem__(self, k: str):
    return self._data[k]

  def __setitem__(self, k: str, v):
    if "." in k:
      k, subk = k.split(".", 1)
      v = {subk: v}
    if isinstance(v, dict):
      v = type(self)(v)
    self._data[k] = v


class Color:
  """Значение цвета"""

  def __init__(self, *args,
               ahex: str = None,
               hex: str = None,
               hexa: str = None,
               rgb: tuple[int, int, int] = None,
               rgba: tuple[int, int, int, int] = None,
               ):
    self._init_complete = False
    self._readonly = False
    self._red = None
    self._green = None
    self._blue = None
    self._alpha = 255
    largs = len(args)
    if largs == 0:
      pass
    elif largs == 3:
      self._check_init()
      self.rgb = args
    elif largs == 4:
      self._check_init()
      self.rgba = args
    else:
      raise TypeError("Give from 3 to 4 positional arguments or 1 keyword argument")
    if not ahex is None:
      self._check_init()
      self.ahex = ahex
    if not hex is None:
      self._check_init()
      self.hex = hex
    if not hexa is None:
      self._check_init()
      self.hexa = hexa
    if not rgb is None:
      self._check_init()
      self.rgb = rgb
    if not rgba is None:
      self._check_init()
      self.rgba = rgba
    if not self._init_complete:
      raise TypeError("Give from 3 to 4 positional arguments or 1 keyword argument")

  def __repr__(self):
    cls = type(self)
    return "%s.%s(rgba=%r)" % (cls.__module__, cls.__name__, self.rgba)

  @property
  def red(self) -> int:
    """Уровень красного от 0 до 255"""
    return self._red

  @red.setter
  def red(self, v):
    self._check_ro()
    self._red = self._check_num(v)
    self._reset()

  @property
  def green(self) -> int:
    """Уровень зелёного от 0 до 255"""
    return self._green

  @green.setter
  def green(self, v):
    self._check_ro()
    self._green = self._check_num(v)
    self._reset()

  @property
  def blue(self) -> int:
    """Уровень синего от 0 до 255"""
    return self._blue

  @blue.setter
  def blue(self, v):
    self._check_ro()
    self._blue = self._check_num(v)
    self._reset()

  @property
  def alpha(self) -> int:
    """Уровень прозрачности от 0 до 255"""
    return self._alpha

  @alpha.setter
  def alpha(self, v):
    self._check_ro()
    self._alpha = self._check_num(v)
    self._reset()

  @property
  def ahex(self) -> str:
    """Цвет в формате AHEX (`aarrggbb`)"""
    if self._ahex is None:
      self._ahex = ("%02x" % self.alpha) + self.hex
    return self._ahex

  @ahex.setter
  def ahex(self, v):
    self.alpha, self.red, self.green, self.blue = self._split_hex(v, 4)

  @property
  def hex(self) -> str:
    """Цвет в формате HEX (`rrggbb`). **Без прозрачности**"""
    if self._hex is None:
      self._hex = "%02x%02x%02x" % self.rgb
    return self._hex

  @hex.setter
  def hex(self, v):
    self.rgb = self._split_hex(v, 3)

  @property
  def hexa(self) -> str:
    """Цвет в формате HEXA (`rrggbbaa`)"""
    if self._hexa is None:
      self._hexa = self.hex + ("%02x" % self.alpha)
    return self._hexa

  @hexa.setter
  def hexa(self, v):
    self.rgba = self._split_hex(v, 4)

  @property
  def rgb(self) -> tuple[int, int, int]:
    """Цвет в формате RGB. **Без прозрачности**"""
    if self._rgb is None:
      self._rgb = (self.red, self.green, self.blue)
    return self._rgb

  @rgb.setter
  def rgb(self, v):
    self.red, self.green, self.blue = v

  @property
  def rgba(self) -> tuple[int, int, int, int]:
    """Цвет в формате RGBA"""
    if self._rgba is None:
      self._rgba = (self.red, self.green, self.blue, self.alpha)
    return self._rgba

  @rgba.setter
  def rgba(self, v):
    self.red, self.green, self.blue, self.alpha = v

  def _check_init(self):
    if self._init_complete:
      raise TypeError("Give from 3 to 4 positional arguments or 1 keyword argument")
    self._init_complete = True

  def _check_num(self, num: int):
    if not isinstance(num, int):
      raise TypeError(type(num), int)
    if (num < 0) or (num > 255):
      raise ValueError("The number should be in the range from 0 to 255, not %i" % num)
    return num

  def _check_ro(self):
    if self._readonly:
      raise RuntimeError("Read-only object")

  def _reset(self):
    self._ahex = None
    self._hex = None
    self._hexa = None
    self._rgb = None
    self._rgba = None

  def _split_hex(self, hex: str, n: int) -> list[int]:
    if hex[0] == "#":
      hex = hex[1:]
    lhex = len(hex)
    if lhex != n * 2:
      raise ValueError("Invalid HEX string")
    nums = [int(num, 16) for num in [hex[i:i + 2] for i in range(0, lhex, 2)]]
    return nums

  @classmethod
  def from_random(cls, red: int | range | slice = None, green: int | range | slice = None, blue: int | range | slice = None, alpha: int | range | slice = 255):
    """Случайный цвет"""
    from random import randint, randrange

    def random(num: int | range | slice):
      if num is None:
        return randint(0, 255)
      if isinstance(num, range) or isinstance(num, slice):
        return randrange(num.start, num.stop, num.step)
      return num
    return cls(random(red), random(green), random(blue), random(alpha))


class _COLORS:
  BLACK = 0, 0, 0
  BLUE = 0, 0, 255
  CYAN = 0, 255, 255
  GREEN = 0, 255, 0
  NULL = 0, 0, 0, 0
  RED = 255, 0, 0
  VIOLET = 255, 0, 255
  WHITE = 255, 255, 255
  YELLOW = 255, 255, 0

  def __init__(self):
    self._colors = {}
    for name in dir(self):
      if name[0] != "_":
        color = Color(*getattr(self, name))
        color._readonly = True
        self._colors[name] = color
        setattr(self, name, color)

  def __contains__(self, k) -> bool:
    return k.upper() in self._colors

  def __getitem__(self, k) -> Color:
    return self._colors[k.upper()]


class AutoaddDict(dict):
  def __init__(self, *args, default_value=None, **kwargs):
    dict.__init__(self, *args, **kwargs)
    self.default_value = default_value

  def __getitem__(self, k):
    if not k in self:
      self[k] = self.default_value
    return dict.__getitem__(self, k)


COLORS = _COLORS()
Error401 = AccessDeniedError
Error403 = AccessDeniedError
Error404 = NotFoundError
Inf = Infinity
NaN = NotANumber



import builtins
from .core import ms
from ._any2json_regs import decoders, encoders, reg_decoder, reg_encoder


def _decode_obj(obj: dict):
  return decoders[obj["type"]](obj["data"])


def _encode_obj(obj) -> dict:
  result = {}
  obj_type = builtins.type(obj)
  if hasattr(obj_type, "__module__"):
    if hasattr(obj_type, "__name__"):
      type = obj_type.__module__ + "." + obj_type.__name__
      if type in encoders:
        result["data"] = encoders[type](obj)
        result["type"] = type
        return result
  for type, func in encoders.items():
    try:
      encoded = func(obj)
      if not encoded is None:
        result["data"] = encoded
        result["type"] = type
        return result
    except Exception:
      pass
  raise ValueError("failed to encode object " + repr(obj))


def decode(text, **kw):
  kw["text"] = text
  return _decode_obj(ms.json.decode(**kw))


def encode(data, **kw) -> str:
  kw["data"] = _encode_obj(data)
  return ms.json.encode(**kw)


def read(path: str, **kw):
  kw["path"] = path
  return _decode_obj(ms.json.read(**kw))


def write(path: str, data, **kw):
  kw["data"] = _encode_obj(data)
  kw["path"] = path
  return ms.json.write(**kw)



import requests
import requests.auth
USER_AGENTS: dict[str, str] = {}
USER_AGENTS["Android"] = "Mozilla/5.0 (Linux; Android 13; K) Chrome/128.0.6613.146"
USER_AGENTS["Android13"] = "Mozilla/5.0 (Linux; Android 13; K) Chrome/128.0.6613.146"


def auth_basic(username: str, password: str, *, client: "BaseClient" = None) -> str:
  """Настроить базовую авторизацию для клиента"""
  if username is None:
    username = client.username
  if password is None:
    password = client.password
  header = requests.auth._basic_auth_str(username, password)
  if not client is None:
    client._headers["Authorization"] = header
  return header


class BaseClient:
  """Базовый клиент для HTTP API"""

  def __init__(self, **kw):
    self._init(**kw)

  def _init(self, session: requests.Session = None):
    self._headers = {}
    self._http = session
    self._params = {}
    self._url = "https://example.com/api/{method}"
    self._url_data = {}

  def __enter__(self):
    return self

  def __exit__(self, *a):
    self.http.close()

  def __repr__(self) -> str:
    cls = type(self)
    return "%s.%s(...)" % (cls.__module__, cls.__name__)

  @property
  def http(self) -> requests.Session:
    """HTTP сессия"""
    if self._http is None:
      self._http = requests.Session()
    return self._http

  def _request(self, http_method: str, api_method: str, *, headers: dict[str, str] = None, params: dict[str, str] = None, raise_for_status: bool = True, url_data: dict[str, str] = None, **kw):
    _headers = self._headers.copy()
    _params = self._params.copy()
    _url_data = self._url_data.copy()
    _url_data["method"] = api_method
    if not headers is None:
      _headers.update(headers)
    if not params is None:
      _params.update(params)
    if not url_data is None:
      _url_data.update(url_data)
    kw["headers"] = _headers
    kw["method"] = http_method
    kw["params"] = _params
    kw["url"] = self._url.format(**_url_data)
    result = self.http.request(**kw)
    if raise_for_status:
      result.raise_for_status()
    return result

  def request(self, *args, **kwargs) -> requests.Response:
    """Отправить запрос к API"""
    return self._request(*args, **kwargs)


class Base(BaseClient):
  def __init_subclass__(cls, **kw):
    from warnings import warn
    warn("Class 'Base' renamed 'BaseClient' starting with version 2.4.6", DeprecationWarning)
    return BaseClient.__init_subclass__(**kw)


class BasicAuthClient(BaseClient):
  """API клиент для авторизации `HTTP Basic`"""

  def _request(self, http_method: str, api_method: str, **kw) -> requests.Response:
    if self._headers.get("Authorization") is None:
      auth_basic(None, None, client=self)
    return BaseClient._request(self, http_method, api_method, **kw)

  @property
  def username(self) -> str:
    """Имя пользователя"""
    return self._username

  @username.setter
  def username(self, value):
    self._headers["Authorization"] = None
    self._username = value

  @property
  def password(self) -> str:
    """Пароль"""
    return self._password

  @password.setter
  def password(self, value):
    self._headers["Authorization"] = None
    self._password = value



"""Работа со словарями"""
from .core import ms
from typing import *
# 2.0.0


def merge(old: dict, new: dict) -> dict:
  """Рекурсиво объединить словари"""
  out = old.copy()
  for k, v in new.items():
    if k in out:
      if isinstance(out[k], dict) and isinstance(v, dict):
        out[k] = merge(out[k], v)
      else:
        out[k] = v
    else:
      out[k] = v
  return out


def reverse(d: dict) -> dict:
  """Развернуть порядок ключей"""
  keys = list(d.keys())[::-1]
  r = {}
  for k in keys:
    r[k] = d[k]
  return r


def sort(d: dict, *, key=None, reverse: bool = False) -> dict:
  """Сортировать порядок ключей"""
  keys = list(d.keys())
  keys.sort(key=key, reverse=reverse)
  r = {}
  for k in keys:
    r[k] = d[k]
  return r


def swap(d: dict) -> dict:
  """Вывернуть словарь (повторяющиеся значения будут перезаписаны)"""
  r = {}
  for k, v in d.items():
    r[v] = k
  return r



"""Работа с процессами"""
import subprocess
from .core import ms
from functools import wraps
from typing import *
Popen_kw = {
    "stderr": subprocess.PIPE,
    "stdin": subprocess.PIPE,
    "stdout": subprocess.PIPE,
}


class Popen(subprocess.Popen):
  def __init__(self, *args, **kw):
    if len(args) == 1:
      if isinstance(args[0], (list, tuple)):
        args = args[0]
    for k, v in Popen_kw.items():
      kw.setdefault(k, v)
    kw["args"] = list(args)
    self.launch_kw = kw
    subprocess.Popen.__init__(self, **kw)

  def force_wait(self):
    """Ожидать завершения процесса игнорируя прерывания"""
    while True:
      try:
        return self.wait()
      except:
        pass

  @ms.utils.riot(daemon=True)
  def wait_on_bg(self):
    """Ожидать завершения процесса в фоне (для Linux)"""
    self.force_wait()


@wraps(subprocess.Popen)
def call(*args, **kw) -> subprocess.Popen:
  """Запустить процесс и ожидать его завершения"""
  p = Popen(*args, **kw)
  p.wait()
  return p
